#pragma once

#include <opencv2/opencv.hpp>


/**
 * @brief Removes noise in the depth image by convolving it with a gaussian filter and restoring the edges afterwards
 *
 */
class FastDepthNoiseRemoval
{
public:
/**
 * @brief Constructs a new Noise Removal Object
 *
 * @param filterSize the size of the gauss filter matrix. the same rules apply as with the opencv gaussianblur function (values should be positive and odd, if zero derived from sigma)
 * @param filterSigma the standart deviation of the gauss filter matrix. the same rules apply as with the opencv gaussianblur function (if it is zero derived from size)
 * @param repairSize the size of the dilation mask that will be applied to the edge restoration mask. this step can be disabled if the value is set < 2;
 * @param repairThreshold the threshold that is used to determine which pixels belong to edges
 */
  FastDepthNoiseRemoval(int filterSize = 3, float filterSigma = 0, int repairSize = 3, float repairThreshold = 0.01);
  /**
   * @brief sets the depth data to use
   *
   * @param depth
   */
  void setInput(const cv::Mat & depth){input_ = depth;}
  cv::Mat getInput() const {return input_;}
  void setFilterSize(int size){filter_size_ = size;}
  void setFilterSigma(float sigma){sigma_ = sigma;}
  void setRepairThreshold(float threshold){repair_threshold_ = threshold;}
  void setRepairSize(int size){repair_size_ = size;}
  void setKillBorder(bool status){kill_border_ = status;}
  void filter(cv::Mat &output);
  cv::Mat1b getMask(){return mask_;}
private:
  cv::Mat input_;
  cv::Mat1b mask_;
  int filter_size_;
  float sigma_;
  float repair_threshold_;
  int repair_size_;
  bool kill_border_;
};

