#include "main.h"
#include <opencv2/opencv.hpp>
#include "fast_depth_noise_removal.hpp"


// We'll be using buffer objects to store the kinect point cloud


struct drawData {
	GLuint vboID;
	GLuint cboID;
	GLuint index;
	GLuint normals;
	GLsizei size;
};


drawData controllerData;

drawData framePoints;
drawData gridData;
drawData exportData;
drawData arrowData;
drawData controllerPosData;

// Intermediate Buffers
unsigned char rgbimage[colorwidth*colorheight*4];    // Stores RGB color image
ColorSpacePoint depth2rgb[width*height];             // Maps depth pixels to rgb pixels
CameraSpacePoint depth2xyz[width*height];			 // Maps depth pixels to 3d coordinates

//buffers of the current frame
glm::vec3 framePositions[width*height];
glm::vec3 frameNormals[width*height];
float frameDepthDifference[width*height];
unsigned char frameColors[width*height * 3];
bool frameValidPoints[width*height];

//export Buffers;
std::vector<glm::vec3> points;
std::vector<unsigned char> colors;

// Kinect Variables
IKinectSensor* sensor;             // Kinect sensor
IMultiSourceFrameReader* reader;   // Kinect data source
ICoordinateMapper* mapper;         // Converts between depth, color, and 3d coordinates
TIMESPAN colorFrameTime = -1;

float KinectSizeX = 0.249f;
float KinectSizeY = 0.042f; // 42 mm, 66mm ist die gesammthhe
float KinectSizeZ = 0.067f;

float KinectOriginOffsetX = ((92 + 73) / 2.0f) / 1000.0f;
float KinectOriginOffsetY = KinectSizeY/2.0f;

glm::vec3 FrameTranslation(0.0,0.0,0.0);



//Vive Tracking
SteamTracking* steamTracking = nullptr;
char* controllerSerial = "LHR-FFFDBD41";
int controllerID = 1;
char* trackerSerial = "LHR-18DC0D92";
int trackerID = 5;
glm::mat4x4* currentControlerPos = nullptr;
glm::mat4x4* currentTrackerPos = nullptr;
bool useTrackerTransform = true;
glm::vec3 lastClicketControlelrPos;
glm::mat4x4 controlerToKinect;
std::vector<glm::vec3> controllerPositions;



bool toggleCam = true;
Camera mainCam;

static bool trackControllerPos = true;

static bool writeFile = false;
static bool captureFrame = false;
int lastX, lastY;



//Filter Settings
float minTrackingKinect = 0.3f;//0.8f;
float maxTrackingKinect = 0.9f;// 2.5f;
	//Custom
//float angleDegrees = 20.0f;
float maxNormalAngle = degreeToRadian(25.0f);
bool enableDepthFilter = false;
bool enableNormalFilter = false;//true;
	//Bilateral / FastDepthNoise
bool enableFastDepthNoiseFilter = false;
float repairThreshold = 100.0f;


bool captureFramesTXT = false;



float degreeToRadian(float degree) {
	return (float) (degree*M_PI / 180.0f);
}

void convertTxtToPnts()
{
		std::vector<glm::vec3> pointsConvert;
		std::vector<unsigned char> colorsConvert;
		
		for (int i = 0; i < 20; i++) {
			std::ifstream myfile;
			char buffer[256];
			 // make sure it's big enough
			snprintf(buffer, sizeof(buffer), "JoinedTowers/pnts_%06d.txt", i);
			//snprintf(buffer, sizeof(buffer), "SeperateTowers/pnts_%06d.txt", i);
			
			myfile.open(buffer, std::ios::in);

			if (!myfile)
			{
				continue;
			}
			std::cout << "Reading File: " << buffer << std::endl;


			glm::vec3 tempvec;
			int r, g, b;
			while (!myfile.eof()) // To get you all the lines.
			{
				myfile >> tempvec.x >> tempvec.y >> tempvec.z >> r >> g >> b;
				pointsConvert.push_back(tempvec);
				colorsConvert.push_back(r);
				colorsConvert.push_back(g);
				colorsConvert.push_back(b);

				//std::cout << tempvec.x << " " << tempvec.y << " " << tempvec.z << " "<< r << " " <<g << " " << b << std::endl;

			}

			myfile.close();
		}

		writePTNS(pointsConvert, colorsConvert, "JoinedTowers/Cloud.pnts");
		//writePTNS(pointsConvert, colorsConvert, "SeperateTowers/Cloud.pnts");

	


}



void DrawCoordinateSystem() {
	//TETAEDRON BY HAND
	glBegin(GL_LINES);

	glColor4f(1.0f, 0.0f, 0.0f, 1.0f);
	glVertex3f(0.0f, 0.0f, 0.0f);
	glVertex3f(1.0f, 0.0f, 0.0f);

	glColor4f(0.0f, 1.0f, 0.0f, 1.0f);
	glVertex3f(0.0f, 0.0f, 0.0f);
	glVertex3f(0.0f, 1.0f, 0.0f);

	glColor4f(0.0f, 0.0f, 1.0f, 1.0f);
	glVertex3f(0.0f, 0.0f, 0.0f);
	glVertex3f(0.0f, 0.0f, 1.0f);

	glEnd();
}

void DrawKinect() {
	glBegin(GL_QUADS);

	//links 82, 5 mm ist origin(siehe bild kinect mesures), kinect v2 : 24, 9 cm x 6, 6 cm x 6, 7cm
	glColor4f(0.5f, 0.5f, 0.5f, 1.0f);
	//front
	glVertex3f(0.0f - KinectOriginOffsetX, 0.0f - KinectOriginOffsetY, 0.0f);
	glVertex3f(KinectSizeX - KinectOriginOffsetX, 0.0f - KinectOriginOffsetY, 0.0f);
	glVertex3f(KinectSizeX - KinectOriginOffsetX, KinectSizeY - KinectOriginOffsetY, 0.0f);
	glVertex3f(0.0f - KinectOriginOffsetX, KinectSizeY - KinectOriginOffsetY, 0.0f);
	 //back
	glVertex3f(0.0f - KinectOriginOffsetX, 0.0f - KinectOriginOffsetY, -KinectSizeZ);
	glVertex3f(KinectSizeX - KinectOriginOffsetX, 0.0f - KinectOriginOffsetY, -KinectSizeZ);
	glVertex3f(KinectSizeX - KinectOriginOffsetX, KinectSizeY - KinectOriginOffsetY, -KinectSizeZ);
	glVertex3f(0.0f - KinectOriginOffsetX, KinectSizeY - KinectOriginOffsetY, -KinectSizeZ);

	//sides
	glVertex3f(0.0f - KinectOriginOffsetX, 0.0f - KinectOriginOffsetY, 0.0f);
	glVertex3f(0.0f - KinectOriginOffsetX, 0.0f - KinectOriginOffsetY, -KinectSizeZ);
	glVertex3f(0.0f - KinectOriginOffsetX, KinectSizeY - KinectOriginOffsetY, -KinectSizeZ);
	glVertex3f(0.0f - KinectOriginOffsetX, KinectSizeY - KinectOriginOffsetY, 0.0f);

	glVertex3f(KinectSizeX - KinectOriginOffsetX, 0.0f - KinectOriginOffsetY, 0.0f);
	glVertex3f(KinectSizeX - KinectOriginOffsetX, 0.0f - KinectOriginOffsetY, -KinectSizeZ);
	glVertex3f(KinectSizeX - KinectOriginOffsetX, KinectSizeY - KinectOriginOffsetY, -KinectSizeZ);
	glVertex3f(KinectSizeX - KinectOriginOffsetX, KinectSizeY - KinectOriginOffsetY, 0.0f);

	//top/bottom

	glVertex3f(0.0f - KinectOriginOffsetX, KinectSizeY - KinectOriginOffsetY, 0.0f);
	glVertex3f(0.0f - KinectOriginOffsetX, KinectSizeY - KinectOriginOffsetY, -KinectSizeZ);
	glVertex3f(KinectSizeX - KinectOriginOffsetX, KinectSizeY - KinectOriginOffsetY, -KinectSizeZ);
	glVertex3f(KinectSizeX - KinectOriginOffsetX, KinectSizeY - KinectOriginOffsetY, 0.0f);

	glVertex3f(0.0f - KinectOriginOffsetX, 0.0f - KinectOriginOffsetY, 0.0f);
	glVertex3f(0.0f - KinectOriginOffsetX, 0.0f - KinectOriginOffsetY, -KinectSizeZ);
	glVertex3f(KinectSizeX - KinectOriginOffsetX, 0.0f - KinectOriginOffsetY, -KinectSizeZ);
	glVertex3f(KinectSizeX - KinectOriginOffsetX, 0.0f - KinectOriginOffsetY, 0.0f);


	glEnd();
}

bool initKinect() {
    if (FAILED(GetDefaultKinectSensor(&sensor))) {
		return false;
	}
	if (sensor) {
		sensor->get_CoordinateMapper(&mapper);

		sensor->Open();
		sensor->OpenMultiSourceFrameReader(
			FrameSourceTypes::FrameSourceTypes_Depth | FrameSourceTypes::FrameSourceTypes_Color,
			&reader);
		return reader?true:false;
	} else {
		return false;
	}
}

void getDepthData(IMultiSourceFrame* frame) {
	IDepthFrame* depthframe;
	IDepthFrameReference* frameref = NULL;
	frame->get_DepthFrameReference(&frameref);
	frameref->AcquireFrame(&depthframe);
	if (frameref) frameref->Release();

	if (!depthframe) return;

	// Get data from frame
	unsigned int sz;
	unsigned short* buf;
	depthframe->AccessUnderlyingBuffer(&sz, &buf);

	//unsigned short filterd[width * height];
	//bilateralFilterOwn(buf, filterd, 5, 12.0, 16.0);
	//const unsigned short* curr = (const unsigned short*)filterd;

	cv::Mat OriginalDepthFrame(height, width, CV_16UC1);// , &buf);
	const unsigned short* curr = (const unsigned short*)buf;

	if (enableFastDepthNoiseFilter) {
		for (unsigned int y = 0; y < height; y++)
		{
			for (unsigned int x = 0; x < width; x++)
			{
				OriginalDepthFrame.ptr<unsigned short>(y)[x] = curr[x + y* width];
			}
		}

		FastDepthNoiseRemoval filter(3, 0.0f, 3, repairThreshold);
		filter.setKillBorder(true);
		filter.setInput(OriginalDepthFrame);
		filter.filter(OriginalDepthFrame);
		//	cv::imshow("Input", OriginalDepthFrame);
		//cv::bilateralFilter(OriginalDepthFrame, FilerdDepth, 15, 80, 80);

		curr = (const unsigned short*)OriginalDepthFrame.data;
	}
	// Write vertex coordinates
	mapper->MapDepthFrameToCameraSpace(width*height, curr, width*height, depth2xyz);

	for (unsigned int y = 0; y < height; y++)
	{
		for (unsigned int x = 0; x < width; x++)
		{
			unsigned int i = x + y* width;
			framePositions[i] = glm::vec3(depth2xyz[i].X, depth2xyz[i].Y, depth2xyz[i].Z);

			if ((enableNormalFilter || enableDepthFilter) && x > 0 && y > 0 && x < width - 1 && y < height - 1) {
				if (enableNormalFilter) {
					//normalen
					float dzdx = (curr[x + 1 + width * y] - curr[x - 1 + width * y]) / 2.0f;
					float dzdy = (curr[x + width * (y + 1)] - curr[x + width * (y - 1)]) / 2.0f;

					glm::vec3 d(-dzdx, -dzdy, 1.0f);
					frameNormals[i] = normalize(d);
				}
				if (enableDepthFilter) {
					// dz tiefenwert
					float z = depth2xyz[i].Z;
					if (true) {
						frameDepthDifference[i] = (depth2xyz[x + 1 + width * (y)].Z - z)*(depth2xyz[x + 1 + width * (y)].Z - z)
							+ (depth2xyz[x - 1 + width * (y)].Z - z)*(depth2xyz[x - 1 + width * (y)].Z - z)
							+ (depth2xyz[x + width * (y + 1)].Z - z)*(depth2xyz[x + width * (y + 1)].Z - z)
							+ (depth2xyz[x + width * (y - 1)].Z - z)*(depth2xyz[x + width * (y - 1)].Z - z);
					}
					else {
						frameDepthDifference[i] = 0.0f;
					}
				}
				//if (x == width / 2 && y == height / 2) std::cout << "Frame Difference " << frameDepthDifference[i]<< std::endl;
			} else {
				frameNormals[i] = glm::vec3(0.0, 0.0, 1.0);
				frameDepthDifference[i] = 0.0f;
			}
			
				
			
		}
	}

	for (unsigned int i = 0; i < sz; i++) {
		
	}

	// Fill in depth2rgb map
	mapper->MapDepthFrameToColorSpace(width*height, buf, width*height, depth2rgb);
	if (depthframe) depthframe->Release();
}

void getRgbData(IMultiSourceFrame* frame) {
	IColorFrame* colorframe;
	IColorFrameReference* frameref = NULL;
	frame->get_ColorFrameReference(&frameref);
	frameref->AcquireFrame(&colorframe);

	TIMESPAN time;
	if (SUCCEEDED(frameref->get_RelativeTime(&time))) {
		colorFrameTime = time;
	}
	else {
		colorFrameTime = -1;
	}
	
	if (frameref) frameref->Release();

	if (!colorframe) return;

	// Get data from frame
	colorframe->CopyConvertedFrameDataToArray(colorwidth*colorheight*4, rgbimage, ColorImageFormat_Rgba);

	// Write color array for vertices
	unsigned char* fdest = (unsigned char*)frameColors;
	bool* fvalid = (bool*)frameValidPoints;

	for (int i = 0; i < width*height; i++) {
		ColorSpacePoint p = depth2rgb[i];
		// Check if color pixel coordinates are in bounds
		if (p.X < 1 || p.Y < 1 || p.X > colorwidth-1 || p.Y > colorheight-1) {
			*fdest++ = 255;
			*fdest++ = 0;
			*fdest++ = 0;
			*fvalid++ = false;
		}
		else {
			

			if (depth2xyz[i].Z > maxTrackingKinect || depth2xyz[i].Z < minTrackingKinect) {
				*fdest++ = 0;
				*fdest++ = 255;
				*fdest++ = 0;

				*fvalid++ = false;
				continue;
			}

			if (enableDepthFilter &&  frameDepthDifference[i] > 100.0) {
					*fdest++ = 255;
					*fdest++ = 140;
					*fdest++ = 0;

					*fvalid++ = false;
					continue;
				}


			if (enableNormalFilter) {
				float angle = dot(frameNormals[i], glm::vec3(0.0, 0.0, 1.0));
				if ( angle < maxNormalAngle) {
					*fdest++ = 0;
					*fdest++ = 0;
					*fdest++ = 255;

					*fvalid++ = false;
					continue;
				}
			}

			int idx = (int)p.X + colorwidth*(int)p.Y;
			*fdest++ = rgbimage[4 * idx + 0];
			*fdest++ = rgbimage[4 * idx + 1];
			*fdest++ = rgbimage[4 * idx + 2];
			*fvalid++ = true;
			
		}
		

		// Don't copy alpha channel
	}

	if (colorframe) colorframe->Release();
}

void getKinectData() {
	IMultiSourceFrame* frame = NULL;

	if (SUCCEEDED(reader->AcquireLatestFrame(&frame))) {
		getDepthData(frame);
		getRgbData(frame);


		glBindBuffer(GL_ARRAY_BUFFER, framePoints.vboID);
		glBufferData(GL_ARRAY_BUFFER, width*height * 3 * sizeof(GLfloat), framePositions, GL_DYNAMIC_DRAW);
		glUnmapBuffer(GL_ARRAY_BUFFER);

		glBindBuffer(GL_ARRAY_BUFFER, framePoints.cboID);
		glBufferData(GL_ARRAY_BUFFER, width*height * 3 * sizeof(unsigned char), frameColors, GL_DYNAMIC_DRAW);

		glUnmapBuffer(GL_ARRAY_BUFFER);
	}
	if (frame) frame->Release();

}

int counter = 0;

void processCurrentFrameForExport() {
	points.reserve(points.size() + width*height);
	colors.reserve(colors.size() + width*height * 3);

	float minx, miny, minz;
	minx = miny = minz = std::numeric_limits<float>::max();
	float maxx, maxy, maxz;
	maxx = maxy = maxz = -std::numeric_limits<float>::max();


	std::ofstream myfile;
	if (captureFramesTXT) {
		char buffer[64]; // make sure it's big enough
		snprintf(buffer, sizeof(buffer), "frames/pnts_%d.txt", counter++);
		myfile.open(buffer, std::ios::out);
	}	
	glm::mat4x4 temp = glm::inverse(*currentTrackerPos);
	for (unsigned int i = 0; i < width*height; i++) {
		if (frameValidPoints[i]) {
			glm::vec3 position = framePositions[i];
			glm::vec3 positionGlobal = glm::vec3(*currentControlerPos * glm::translate(glm::mat4(), FrameTranslation) * controlerToKinect * glm::vec4(position, 1.0f));

			if (isinf(position.x) || isinf(position.y) || isinf(position.z))
				continue;
			
			if (position.x < minx) minx = position.x;
			if (position.x > maxx) maxx = position.x;
			if (position.y < miny) miny = position.y;
			if (position.y > maxy) maxy = position.y;
			if (position.z < minz) minz = position.z;
			if (position.z > maxz) maxz = position.z;

			points.push_back(positionGlobal);
			colors.push_back(frameColors[i * 3]);
			colors.push_back(frameColors[i * 3 + 1]);
			colors.push_back(frameColors[i * 3 + 2]);

			if (captureFramesTXT) {
				glm::vec3 tempvec = glm::vec3(temp* glm::vec4(positionGlobal, 1.0f));
				myfile << tempvec.x << " " << tempvec.y << " " << tempvec.z << " "
					<< (int)frameColors[i * 3] << " " << (int)frameColors[i * 3 + 1] << " " << (int)frameColors[i * 3 + 2] << std::endl;
			}

		}
	}

	if (captureFramesTXT) {
		myfile.close();
	}
	std::cout << "Frame Local Boundingbox Min: [" << minx << "|" << miny << "|" << minz << "] Max: [" << maxx << "|" << maxy << "|" << maxz << "]"<< std::endl;
	//update Buffers
	if (points.size() == 0) return;

	glBindBuffer(GL_ARRAY_BUFFER, exportData.vboID);
	glBufferData(GL_ARRAY_BUFFER, points.size() * sizeof(glm::vec3), &points[0], GL_DYNAMIC_DRAW);
	glUnmapBuffer(GL_ARRAY_BUFFER);
	
	glBindBuffer(GL_ARRAY_BUFFER, exportData.cboID);
	glBufferData(GL_ARRAY_BUFFER, colors.size() * sizeof(unsigned char), &colors[0], GL_DYNAMIC_DRAW);
	exportData.size = points.size();

	glUnmapBuffer(GL_ARRAY_BUFFER);

}

void writePTNS(std::vector<glm::vec3> exportpoints, std::vector<unsigned char> exportColors, const char* filename) {
	pntsHeader header;
	size_t pointArrayByteSize = exportpoints.size() * sizeof(glm::vec3);
	std::cout << "writing File: Number of Points:" << exportpoints.size();
	size_t coloArrayByteSize = exportColors.size() * sizeof(unsigned char);
	header.featureTableByteLenght = pointArrayByteSize + coloArrayByteSize;

	std::ostringstream os;
	os << "{\"POINTS_LENGTH\":" << (exportpoints.size()) << ", \"POSITION\" : {\"byteOffset\":0}, \"RGB\" : {\"byteOffset\":" << pointArrayByteSize << "}}";
	std::string json = os.str();

	header.featureTableJSONByteLenght = json.size();
	header.byteLenght = 28 /*header*/ + header.featureTableJSONByteLenght + header.featureTableByteLenght;
	std::cout << " File Size in byte" << header.byteLenght << std::endl;

	std::ofstream myfile;
	myfile.open(filename, std::ios::out | std::ios::binary);
	myfile.write((char*)&header, 28);
	myfile << json;
	myfile.write(reinterpret_cast<const char*>(&exportpoints[0]), pointArrayByteSize);
	myfile.write(reinterpret_cast<const char*>(&exportColors[0]), coloArrayByteSize);
	//myfile << 


	myfile.close();

}

void writePointCloud() {
	if (points.size() == 0) {
		std::cout << "No points, scipping File write" << std::endl;
		return;
	}

	currentTrackerPos = steamTracking->getTransformationForDevice(trackerID, 0.0f);
	currentTrackerPos = new glm::mat4x4(glm::rotate(*currentTrackerPos, degreeToRadian(-90), glm::vec3(1.0f, 0.0f, 0.0f)));

	std::vector<glm::vec3> exportpoints;
	exportpoints.reserve(points.size());
	for (size_t i = 0; i < points.size(); i++) {
		

		exportpoints.push_back(glm::inverse(*currentTrackerPos) * glm::vec4(points[i], 1.0f));
	}

	//exportpoints.clear();
	/*
	for (int i = 0; i < 1000; i++) {
		exportpoints.push_back(glm::vec3(i/1000.0f, 0.0f, 0.0f));
		exportpoints.push_back(glm::vec3(0.0f, i / 1000.0f, 0.0f));
		//exportpoints.push_back(glm::vec3(0.0f, i / 2000.0f, 1.0f));
		exportpoints.push_back(glm::vec3(0.0f, 0.0f, i / 1000.0f));
	}*/

	writePTNS(exportpoints, colors, "example.pnts");

}



void rotateCamera() {
	static double angle = 0.;
	static double radius = 20.;
	double x = radius*sin(angle);
	double z = radius*(1-cos(angle))-radius/2;
	
	//gluLookAt(0, 0, 0,		0, 5, 0, 	0, 1, 0);

	gluLookAt(x,10,z, 0,0,0, 0,1,0);
	angle += 0.002;
}

void drawElementArray(drawData data, GLenum mode) {
	glBindBuffer(GL_ARRAY_BUFFER, data.vboID);
	glVertexPointer(3, GL_FLOAT, 0, NULL);

	glBindBuffer(GL_ARRAY_BUFFER, data.cboID);
	glColorPointer(3, GL_UNSIGNED_BYTE, 0, NULL);

	glDrawArrays(mode, 0, data.size);
}

void mainRenderLoop() {
	steamTracking->processEvents();

	mainCam.tick();

	getKinectData();
	 
	time_t currentTime;

	time(&currentTime);
	
	//std::cout << "Relative Time " << colorFrameTime << " current "<< currentTime << " dif: " << (currentTime - colorFrameTime/10) <<  std::endl;
	//get controler pos
	delete currentControlerPos;
	currentControlerPos = steamTracking->getTransformationForDevice(controllerID, 0.0f);
	delete currentTrackerPos;
	currentTrackerPos = steamTracking->getTransformationForDevice(trackerID, 0.0f);
	currentTrackerPos = new glm::mat4x4(glm::rotate(*currentTrackerPos, degreeToRadian(-90), glm::vec3(1.0f, 0.0f, 0.0f)));

	

	if (trackControllerPos) {
		glm::vec3 tempPos = glm::vec3(*currentControlerPos*glm::vec4(0, 0, 0, 1));
		controllerPositions.push_back(tempPos);

		if (controllerPositions.size() > 2) {

			glBindBuffer(GL_ARRAY_BUFFER, controllerPosData.vboID);
			glBufferData(GL_ARRAY_BUFFER, controllerPositions.size() * sizeof(glm::vec3), &controllerPositions[0], GL_DYNAMIC_DRAW);
			controllerPosData.size = controllerPositions.size();

		}
	}





	//add points to point cloud
	if (captureFrame) {
		processCurrentFrameForExport();
		
		glm::vec3 TrackerPos = glm::vec3(*steamTracking->getTransformationForDevice(trackerID, 0.0f)*glm::vec4(0, 0, 0, 1));

		glm::vec3 current = controllerPositions.back();
		std::cout << "Current Position [" << current.x << "/" << current.y << "/" << current.z << "] TrackerPos: [" << TrackerPos.x << "/" << TrackerPos.y << "/" << TrackerPos.z << "] distance " << distance(lastClicketControlelrPos, current) << std::endl ;
		lastClicketControlelrPos = controllerPositions.back();

		FrameTranslation = glm::vec3(0.0, 0.0, 0.0);
		captureFrame = false;
	}

	if (writeFile) {
		writePointCloud();
		writeFile = false;
	}

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	//rotateCamera(); //todo
	if (toggleCam ) {

		glTranslatef(0.0f, -0.3f, -1.0f);
		glRotatef(10.0, 1.0, 0.0, 0.0);
		glRotatef(180.0, 0.0, 1.0, 0.0);
		glMultMatrixf(glm::value_ptr(glm::inverse(controlerToKinect)));
		glMultMatrixf(glm::value_ptr(glm::inverse(*currentControlerPos)));
		//
	}
	else {
		mainCam.applyMatrix();
	}

	//glTranslatef(0.0, 5, 0.0);

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_COLOR_ARRAY);


	//draw Grid
	glLineWidth(1.f);
	drawElementArray(gridData, GL_LINES);

	//draw Export
	drawElementArray(exportData, GL_POINTS);

	glDisableClientState(GL_COLOR_ARRAY);

	//draw Controller Positions
	
	if (trackControllerPos && controllerPosData.size > 1){

		glColor4f(0.0f, 0.0f, 1.0f, 1.0f);
		glBindBuffer(GL_ARRAY_BUFFER, controllerPosData.vboID);
		glVertexPointer(3, GL_FLOAT, 0, NULL);
		glDrawArrays(GL_LINE_STRIP, 0, controllerPosData.size);
	}
	glPushMatrix();
	glMultMatrixf(glm::value_ptr(*currentTrackerPos));
	DrawCoordinateSystem();
	glPopMatrix();

	{ // Controller Space
		glPushMatrix(); 


		if (currentControlerPos) { //applay controler Transformation
			glMultMatrixf(glm::value_ptr(*currentControlerPos));
		}
		DrawCoordinateSystem();

		{// draw controller
			glColor4f(0.5f, 0.5f, 0.5f, 1.0f);

			glBindBuffer(GL_ARRAY_BUFFER, controllerData.vboID);
			glVertexPointer(3, GL_FLOAT, 0, NULL);

			glBindBuffer(GL_ARRAY_BUFFER, controllerData.normals);
			glNormalPointer(GL_FLOAT, 0, NULL);

			glDrawElements(GL_TRIANGLES, controllerData.size, GL_UNSIGNED_SHORT, 0);
		}

		{//Kinect Space
			glPushMatrix(); 
			glTranslatef(FrameTranslation.x, FrameTranslation.y, FrameTranslation.z);
			glMultMatrixf(glm::value_ptr(controlerToKinect));
			
			DrawCoordinateSystem();

			glColor4f(0.2f, 0.2f, 0.2f, 1.0f);
			DrawKinect();

			glEnableClientState(GL_COLOR_ARRAY);
			//draw Frame Point cloud	
			glPointSize(1.f);
			drawElementArray(framePoints, GL_POINTS);

			glPopMatrix();
		}
		glPopMatrix();
	}

	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_COLOR_ARRAY);

}

void moveCurrentFrame(glm::vec3 dir) {
	float factor = 0.002f;
	FrameTranslation += factor*dir;
	std::cout << "Frame Transformation ";
	printVector(FrameTranslation);
	std::cout << std::endl;

}

void keyPressed(unsigned char key, int x, int y)
{
	mainCam.keyPressed(key, x, y);
	switch (key) {
	case 'y':
		std::cout << "Capturing Frame Keyboard ";
		captureFrame = true;
		break;
	case 'x':
		writeFile = true;
		break;
	case 'c':
		points.clear();
		colors.clear();
		exportData.size = 0;
		break;
	case 'q':
		toggleCam = !toggleCam;
		break;
	case 'u':
		moveCurrentFrame(glm::vec3(1.0, 0.0, 0.0));
		break;
	case 'j':
		moveCurrentFrame(glm::vec3(-1.0, 0.0, 0.0));
		break;
	case 'i':
		moveCurrentFrame(glm::vec3(0.0, 1.0, 0.0));
		break;
	case 'k':
		moveCurrentFrame(glm::vec3(0.0, -1.0, 0.0));
		break;
	case 'o':
		moveCurrentFrame(glm::vec3(0.0, 0.0, 1.0));
		break;
	case 'l':
		moveCurrentFrame(glm::vec3(0.0, 0.0, -1.0));
		break;
	case 'p':
		FrameTranslation = glm::vec3(0.0, 0.0, 0.0);
	case 'b':
		convertTxtToPnts();

		break;
	case 27: // Escape key
		exit(0);
		break;
	default:
		break;
	}
	glutPostRedisplay(); /* this redraws the scene without
						 waiting for the display callback so that any changes appear
						 instantly */
}

void keyReleased(unsigned char key, int x, int y) {
	mainCam.keyReleased(key, x, y);
}

void mouseButton(int button, int state, int x, int y) {
	mainCam.mouseButton(button, state, x, y);
}

void mouseMove(int x, int y) {
	mainCam.mouseMove(x, y);
}

void captureNextFrame(vr::VREvent_t event)
{
	if (event.trackedDeviceIndex == controllerID && event.data.controller.button == 33) {
		std::cout << "Capturing Frame Controler Button ";
		captureFrame = true;
	}
}




void printVector(glm::vec3 vector)
{
	std::cout << "[" << vector.x << "/" << vector.y << "/" << vector.z << "]";
}


int main(int argc, char* argv[]) {
	std::cout << "Starting Kinect VR Point Cloud creator\n";
    if (!init(argc, argv)) return 1;
    if (!initKinect()) return 1;

	steamTracking = new SteamTracking();
	steamTracking->printVRDevices();
	steamTracking->onButtonPress = captureNextFrame;
	controllerID = steamTracking->getDeviceIDForSerial(controllerSerial);
	trackerID = steamTracking->getDeviceIDForSerial(trackerSerial);


    // OpenGL setup
    glClearColor(0,0,0,0);
    glClearDepth(1.0f);

	setupFrameBuffer();

	setupGrid();

	glGenBuffers(1, &exportData.vboID);
	glGenBuffers(1, &exportData.cboID);
	exportData.size = 0;

	glUnmapBuffer(GL_ARRAY_BUFFER);

    // Camera setup
    glViewport(0, 0, width, height);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
	gluPerspective(45, width /(GLdouble) height, 0.1, 1000);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

	//kienct: https://msdn.microsoft.com/de-de/library/dn785530.aspx
	// 1 unit = 1 meter
	// z ist nach vorne, y up, x links aka -90 um die y achse gedreht
	// The Camera Space origin (x=0, y=0, z=0) is located at the center of the IR sensor on Kinect 
	//von links 82,5 mm ist origin (siehe bild kinect mesures),  kinect v2: 24,9 cm x 6,6 cm x 6,7cm
	//


	// http://www.opengl-tutorial.org/beginners-tutorials/tutorial-7-model-loading/
	//C:\Program Files (x86)\Steam\steamapps\common\SteamVR\resources\rendermodels\vr_controller_vive_1_5
	//HKEY_CURRENT_USER/Software/Valve/SteamPath.



	//controlerToKinect = glm::translate(glm::mat4x4(), (glm::vec3(-4.0f, -7.5f, 13.0f)*0.01f));


	
	//controlerToKinect = glm::translate(glm::mat4x4(), (glm::vec3(0.0f, 0.0f,0.173 /*controller*/+0.03 /*Halterung*/ )));
	controlerToKinect = glm::mat4x4();
	controlerToKinect = glm::rotate(controlerToKinect, degreeToRadian(-58+180 /*63,9...*/), glm::vec3(1.0f, 0.0f, 0.0f));
	controlerToKinect = glm::rotate(controlerToKinect, degreeToRadian(180.0f), glm::vec3(0.0f, 1.0f, 0.0f));
	//controlerToKinect = glm::rotate(controlerToKinect, degreeToRadioan(180.0f), glm::vec3(0.0f, 0.0f, 1.0f));

	float dx = 0.000f;// 0.01;
	float dy = KinectSizeY / 2.0f + 0.004f /*HAlterung Platte*/ + 0.011f /* Conroller*/;
	float dz = -0.01f; //-0.023 + 0.02; // 0.045f /*= versatz auf der Halterung; Distanz zur Halterung:*/ + (KinectSizeZ - (0.066 /*halterung z*/ - 0.012 /*halterung berstand*/));

	controlerToKinect = glm::translate(controlerToKinect, (glm::vec3(dx, -dy, dz)));


	glm::vec3 tempPos = glm::vec3(controlerToKinect*glm::vec4(0, 0, 0, 1));
	std::cout << "Controller Kinect Transformation ";
		printVector(tempPos);
	std::cout << std::endl;
	std::cout << "Kinect to center: " << dx <<" "<< dy << " " << dz <<std::endl;
	
	// Error Testing
	glm::mat4 kinectToController = glm::inverse(controlerToKinect);

	float angleInDegrees = 30;
	glm::vec4 test = controlerToKinect * glm::rotate(glm::mat4(), degreeToRadian(angleInDegrees), glm::vec3(1.0f, 0.0f, 0.0f)) * kinectToController * glm::vec4(0.0, 0.0, 1.0, 1.0);
	std::cout << "Angle Error x ";
	printVector(test);
	std::cout << std::endl;

	test = controlerToKinect * glm::rotate(glm::mat4(), degreeToRadian(angleInDegrees), glm::vec3(0.0f, 1.0f, 0.0f)) * kinectToController * glm::vec4(0.0, 0.0, 1.0, 1.0);
	std::cout << "Angle Error y ";
	printVector(test);
	std::cout << std::endl;

	test = controlerToKinect * glm::rotate(glm::mat4(), degreeToRadian(angleInDegrees), glm::vec3(0.0f, 0.0f, 1.0f)) * kinectToController * glm::vec4(0.0, 0.0, 1.0, 1.0);
	std::cout << "Angle Error z ";
	printVector(test);
	std::cout << std::endl;


	//
	std::vector<unsigned short> indices;
	std::vector<glm::vec3> out_vertices;
	std::vector<glm::vec2> out_uvs;
	std::vector<glm::vec3> out_normals;


	loadAssImp("C:\\Program Files (x86)\\Steam\\steamapps\\common\\SteamVR\\resources\\rendermodels\\vr_controller_vive_1_5\\vr_controller_vive_1_5.obj", indices, out_vertices, out_uvs, out_normals);

	float max = 0.0;
	for (size_t i = 0; i < out_vertices.size(); i++)
	{
		max = std::fmax(out_vertices[i].z, max);
	}
	std::cout << "max " << max << std::endl;

	glGenBuffers(1, &controllerData.vboID);
	glBindBuffer(GL_ARRAY_BUFFER, controllerData.vboID);
	glBufferData(GL_ARRAY_BUFFER, out_vertices.size() * sizeof(glm::vec3), &out_vertices[0], GL_STATIC_DRAW);

	glGenBuffers(1, &controllerData.normals);
	glBindBuffer(GL_ARRAY_BUFFER, controllerData.normals);
	glBufferData(GL_ARRAY_BUFFER, out_normals.size() * sizeof(glm::vec3), &out_normals[0], GL_STATIC_DRAW);

	glGenBuffers(1, &controllerData.index);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, controllerData.index);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, indices.size() * sizeof(unsigned short), &indices[0], GL_STATIC_DRAW);
	controllerData.size = indices.size();

	controllerPositions.clear();
	glGenBuffers(1, &controllerPosData.vboID);
	glBindBuffer(GL_ARRAY_BUFFER, controllerPosData.vboID);

	glBufferData(GL_ARRAY_BUFFER, 0 * sizeof(glm::vec3), &out_vertices[0], GL_DYNAMIC_DRAW);
	controllerPosData.size = 0;

    // Main loop
	atexit(close);
    execute();
	
    return 0;
}

void setupFrameBuffer()
{
	{ // init depth Image
	  // Write color array for vertices
		unsigned char* fcolor = (unsigned char*)frameColors;
		glm::vec3* fvec3 = (glm::vec3*)framePositions;
		for (int j = 0; j < height; j++) {
			for (int i = 0; i < width; i++) {
				*fvec3++ = glm::vec3(1.0 * i / width, 1.0 *j / height, 0);
				*fcolor++ = (unsigned char)(1.0 * i / width * 255);
				*fcolor++ = (unsigned char)(1.0 * j / height * 255);
				*fcolor++ = 0;
				frameValidPoints[i + width*j] = true;
			}
		}


		// Set up array buffers
		glGenBuffers(1, &framePoints.vboID);
		glBindBuffer(GL_ARRAY_BUFFER, framePoints.vboID);
		glBufferData(GL_ARRAY_BUFFER, width*height * sizeof(glm::vec3), framePositions, GL_DYNAMIC_DRAW);
		glGenBuffers(1, &framePoints.cboID);
		glBindBuffer(GL_ARRAY_BUFFER, framePoints.cboID);
		glBufferData(GL_ARRAY_BUFFER, width*height * 3 * sizeof(unsigned char), frameColors, GL_DYNAMIC_DRAW);
		framePoints.size = width*height;
	}
}

void setupGrid()
{
	{ //init Grid
		std::vector<glm::vec3> grid;
		std::vector<unsigned char> gridcolors;
		for (int i = -20; i < 21; i++)
		{
			grid.push_back(glm::vec3(20, 0, i));
			grid.push_back(glm::vec3(-20, 0, i));
			
			gridcolors.push_back(i == 0 ? 0 : 255);
			gridcolors.push_back(i == 0 ? 0 : 255);
			gridcolors.push_back(255);

			
			gridcolors.push_back(i == 0 ? 0 : 255);
			gridcolors.push_back(i == 0 ? 0 : 255);
			gridcolors.push_back(255);

		}
		for (int i = -20; i < 21; i++)
		{
			grid.push_back(glm::vec3(i, 0, 20));
			grid.push_back(glm::vec3(i, 0, -20));
			gridcolors.push_back(255);
			gridcolors.push_back(i == 0 ? 0 : 255);
			gridcolors.push_back(i == 0 ? 0 : 255);

			gridcolors.push_back(255);
			gridcolors.push_back(i == 0 ? 0 : 255);
			gridcolors.push_back(i == 0 ? 0 : 255);
		}
		grid.push_back(glm::vec3(0, -20, 0));
		grid.push_back(glm::vec3(0, 20, 0));
		
		gridcolors.push_back(0);
		gridcolors.push_back(255);
		gridcolors.push_back(0);

		gridcolors.push_back(0);
		gridcolors.push_back(255);
		gridcolors.push_back(0);

		glGenBuffers(1, &gridData.vboID);
		glBindBuffer(GL_ARRAY_BUFFER, gridData.vboID);
		glBufferData(GL_ARRAY_BUFFER, grid.size() * sizeof(glm::vec3), &grid[0], GL_STATIC_DRAW);

		glGenBuffers(1, &gridData.cboID);
		glBindBuffer(GL_ARRAY_BUFFER, gridData.cboID);
		glBufferData(GL_ARRAY_BUFFER, gridcolors.size() * sizeof(unsigned char), &gridcolors[0], GL_STATIC_DRAW);
		gridData.size = grid.size();

	}
}

void close() {
	delete steamTracking;
}

void reshapeFunction(int width, int height) {
	glViewport(0, 0, width, height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(45, width / (GLdouble)height, 0.1, 1000);
	glMatrixMode(GL_MODELVIEW);

}
